﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ArcGIS.Core.CIM;
using ArcGIS.Core.Data;
using ArcGIS.Core.Geometry;
using ArcGIS.Core.SystemCore;
using ArcGIS.Desktop.Catalog;
using ArcGIS.Desktop.Core;
using ArcGIS.Desktop.Editing;
using ArcGIS.Desktop.Extensions;
using ArcGIS.Desktop.Framework;
using ArcGIS.Desktop.Framework.Contracts;
using ArcGIS.Desktop.Framework.Dialogs;
using ArcGIS.Desktop.Framework.Threading.Tasks;
using ArcGIS.Desktop.Mapping;

namespace CoordinateGeometrySDK
{
  internal class CreateCircularArc : Button
  {
    protected override void OnClick()
    {
      //Check for one selected layer
      if (MapView.Active.GetSelectedLayers().OfType<FeatureLayer>().Count() != 1)
      {
        System.Windows.MessageBox.Show("Please select one line feature layer in the table of contents", "Construct Circular Arc");
        return;
      }

      var featLyr = MapView.Active.GetSelectedLayers().OfType<FeatureLayer>().First();

      //Run on MCT
      QueuedTask.Run( () =>
      {
        #region get polyline feature layer
        var fcDefinition = featLyr?.GetFeatureClass()?.GetDefinition();
        if (fcDefinition == null)
          return;

        if (fcDefinition.GetShapeType() != GeometryType.Polyline)
        {
          System.Windows.MessageBox.Show("Please select a line feature layer in the table of contents", "Construct Circular Arc");
          return;
        }
        #endregion

        //Example scenario. Make a circular arc that has 30 meter radius and 60 meter chord, N10°E chord bearing
        //=====================================================
        //User Entry Values
        //=====================================================
        double dRadius = 60;  //in meters
        double dChord = 60; //in meters
        string QuadrantBearingChord = "n10-00-0e";
        bool curveLeft = true; //curve turning towards the left when traveling from start point towards end point.
        //=====================================================

        double dMetersPerUnit = 1;
        if (fcDefinition.GetSpatialReference().IsProjected)
          dMetersPerUnit = fcDefinition.GetSpatialReference().Unit.ConversionFactor;

        //We know the incoming length values are already metric, but we don’t know what the user’s target linear unit is.
        //We can use the metric converter to get the value for the target linear unit
        //These values are the "ground" Radius and Chord values

        dRadius /= dMetersPerUnit;  //60 meters divided by meters per unit
        dChord /= dMetersPerUnit;

        double dChordBearing = QuadrantBearingDMSToPolarRadians(QuadrantBearingChord);

        ArcOrientation CCW = curveLeft ? ArcOrientation.ArcCounterClockwise : ArcOrientation.ArcClockwise;

        #region get the ground to grid corrections
        //Get the active map view.
        var mapView = MapView.Active;
        if (mapView?.Map == null)
          return;

        var cimDefinition = mapView.Map?.GetDefinition();
        if (cimDefinition == null) return;
        var cimG2G = cimDefinition.GroundToGridCorrection;

        //These extension methods automatically check if ground to grid is active, etc.
        double dG2G_ScaleFactor = cimG2G.GetConstantScaleFactor();
        double dG2G_DirectionOffsetCorrection = cimG2G.GetDirectionOffset() * Math.PI / 180;
        //this property is in decimal degrees. Converted to radians for use in circular arc creation

        #endregion

        double dGridRadius = dRadius * dG2G_ScaleFactor;
        double dGridChord = dChord * dG2G_ScaleFactor;
        double dGridChordBearing = dChordBearing + dG2G_DirectionOffsetCorrection;
        var circArcStartPoint = MapView.Active.Extent.Center;
        if (circArcStartPoint == null)
          return;

        //Create the circular arc geometry
        EllipticArcSegment pCircArc = null;

        try
        {
          pCircArc = EllipticArcBuilderEx.CreateCircularArc(circArcStartPoint, dGridChord, 
            dGridChordBearing, dGridRadius, CCW, MinorOrMajor.Minor);
        }
        catch
        {
          System.Windows.MessageBox.Show("Circular arc parameters not valid.", "Construct Circular Arc");
          return;
        }

        var newPolyline = PolylineBuilderEx.CreatePolyline(pCircArc);

        Dictionary<string, object> MyAttributes = new Dictionary<string, object>();
        MyAttributes.Add(fcDefinition.GetShapeField(), newPolyline);

        //check to make sure line is COGO enabled
        if (fcDefinition.IsCOGOEnabled())
        {
          //storing the entered direction in north azimuth decimal degrees
          MyAttributes.Add("Direction", PolarRadiansToNorthAzimuthDecimalDegrees(dChordBearing));

          if (curveLeft)
            dRadius = -dRadius; //store curves to the left with negative radius.

          MyAttributes.Add("Radius", dRadius);
          //the arclength on the geometry is grid, so convert to ground
          MyAttributes.Add("Arclength", pCircArc.Length / dG2G_ScaleFactor);
        }

        var op = new EditOperation
        {
          Name = "Construct Circular Arc",
          SelectNewFeatures = true
        };
        op.Create(featLyr, MyAttributes);
        op.Execute();
      });
    }

    private double QuadrantBearingDMSToPolarRadians(string InQuadBearingDMS)
    {
      var AngConv = DirectionUnitFormatConversion.Instance;
      var ConvDef = new ConversionDefinition()
      {
        DirectionTypeIn = ArcGIS.Core.SystemCore.DirectionType.QuadrantBearing,
        DirectionUnitsIn = ArcGIS.Core.SystemCore.DirectionUnits.DegreesMinutesSeconds,
        DirectionTypeOut = ArcGIS.Core.SystemCore.DirectionType.Polar,
        DirectionUnitsOut = ArcGIS.Core.SystemCore.DirectionUnits.Radians
      };
      return AngConv.ConvertToDouble(InQuadBearingDMS, ConvDef);
    }

    private double PolarRadiansToNorthAzimuthDecimalDegrees(double InPolarRadians)
    {
      var AngConv = DirectionUnitFormatConversion.Instance;
      var ConvDef = new ConversionDefinition()
      {
        DirectionTypeIn = ArcGIS.Core.SystemCore.DirectionType.Polar,
        DirectionUnitsIn = ArcGIS.Core.SystemCore.DirectionUnits.Radians,
        DirectionTypeOut = ArcGIS.Core.SystemCore.DirectionType.NorthAzimuth,
        DirectionUnitsOut = ArcGIS.Core.SystemCore.DirectionUnits.DecimalDegrees
      };
      return AngConv.ConvertToDouble(InPolarRadians, ConvDef);
    }
  }
}
